# Test convergence for biharmonic problem
from firedrake import *
from firedrake.petsc import PETSc
from mpi4py import MPI
import numpy
import os

if COMM_WORLD.rank == 0:
    if not os.path.exists("data"):
        os.makedirs("data")
    elif not os.path.isdir("data"):
        raise RuntimeError("Cannot create output directory, file of given name exists")
COMM_WORLD.barrier()

els = ["Lagrange"]*4
degs = range(2, 6)


def uexact(x, y):
    return Constant(100) * (x*(1-x)*y*(1-y))**2


def delta(u):
    return div(grad(u))


def dds(u, n):
    if isinstance(n, list) or isinstance(n, tuple):
        expr = u
        for s in n:
            expr = dot(grad(expr), s)
        return expr
    else:
        return dot(grad(u), n)


PETSc.Log().begin()


def get_time(event, comm=COMM_WORLD):
    return comm.allreduce(PETSc.Log.Event(event).getPerfInfo()["time"], op=MPI.SUM) / comm.size


N_base = 8
mesh = UnitSquareMesh(N_base, N_base)
mh = MeshHierarchy(mesh, 4)
# Let's perturb the original mesh
V = FunctionSpace(mesh, mesh.coordinates.ufl_element())
eps = Constant(3 / 2**(N_base-1))

x, y = SpatialCoordinate(mesh)
new = Function(V).interpolate(as_vector([x + eps*sin(2*pi*x)*sin(2*pi*y),
                                         y - eps*sin(2*pi*x)*sin(2*pi*y)]))

# And propagate to refined meshes
coords = [new]
for mesh in mh[1:]:
    fine = Function(mesh.coordinates.function_space())
    prolong(new, fine)
    coords.append(fine)
    new = fine

for mesh, coord in zip(mh, coords):
    mesh.coordinates.assign(coord)

for el, deg in zip(els, degs):
    PETSc.Sys.Print("{}-{}".format(el, deg))
    results = []
    element = FiniteElement(el, triangle, deg)
    for i, msh in enumerate(mh):
        N = N_base * 2**i
        x, y = SpatialCoordinate(msh)
        uex = uexact(x, y)

        f = delta(delta(uex))

        V = FunctionSpace(msh, element)
        bc = DirichletBC(V, Constant(0), "on_boundary")

        n = FacetNormal(msh)
        t = dot(as_tensor([[0, -1], [1, 0]]), n)
        h = CellSize(msh)
        h_avg = (h('+') + h('-'))/2.0
        u = TrialFunction(V)
        v = TestFunction(V)

        # Need larger penalty parameter for P5 to converge.
        # alpha = 50 works down to 64x64, 128x128 runs out of
        # precision?
        alpha = Constant(50.0)
        u = TrialFunction(V)
        v = TestFunction(V)
        beta = Constant(40)

        a = (inner(div(grad(u)), div(grad(v)))*dx -
             inner(avg(div(grad(u))), jump(grad(v), n))*dS -
             inner(jump(grad(u), n), avg(div(grad(v))))*dS +
             alpha/h_avg*inner(jump(grad(u), n), jump(grad(v), n))*dS -
             inner(delta(u), dds(v, n))*ds -
             inner(dds(u, n), delta(v))*ds +
             beta/h * inner(dds(u, n), dds(v, n))*ds)
        L = inner(f, v)*dx

        uh = Function(V)

        params = {"snes_type": "newtonls",
                  "snes_linesearch_type": "basic",
                  "snes_lag_jacobian": -2,
                  "snes_lag_preconditioner": -2,
                  "snes_max_it": 2,
                  "ksp_type": "preonly",
                  "pc_type": "lu",
                  "pc_factor_mat_solver_type": "pastix",
                  "snes_rtol": 1e-16,
                  "snes_atol": 1e-25}

        if i == 0:
            # Warm up symbolics/disk cache
            try:
                solve(a == L, uh, bcs=bc, solver_parameters=params)
            except ConvergenceError:
                pass

        # in case iterative refinement doesn't work
        with PETSc.Log.Stage("{el}{deg}.N{N}".format(el=el, deg=deg, N=N)):
            try:
                solve(a == L, uh, bcs=bc, solver_parameters=params, options_prefix="")
            except ConvergenceError:
                pass
            snes = get_time("SNESSolve")
            ksp = get_time("KSPSolve")
            pcsetup = get_time("PCSetUp")
            pcapply = get_time("PCApply")
            jac = get_time("SNESJacobianEval")
            residual = get_time("SNESFunctionEval")
            sparsity = get_time("CreateSparsity")

        error = sqrt(assemble((uex-uh)**2*dx))

        results.append([N, error, snes, ksp, pcsetup, pcapply, jac, residual, sparsity])
    if mesh.comm.rank == 0:
        with open('data/biharmonicdg.{el}.{deg}.csv'.format(el=el, deg=deg), 'w') as f:
            numpy.savetxt(f, results, fmt=['%d'] + ['%e'] * 8, delimiter=',',
                          header='N,Error,SNESSolve,KSPSolve,PCSetUp,PCApply,SNESJacobianEval,SNESFunctionEval,CreateSparsity', comments='')
