# example usage
# python guitar.py -eps_nev 5 -eps_tol 1e-11 -eps_target 10 -st_type sinvert
from firedrake import *
from petsc4py import PETSc
from slepc4py import SLEPc
import numpy as np

# Boundary 1: external
# Boundary 2: sound hole
msh = Mesh("cutaway.msh")
# Some finer meshes for visualisation
mh = MeshHierarchy(msh, 1)
vizmesh = mh[-1]
deg = 2
V = FunctionSpace(msh, "Morley", deg)
Pk = FunctionSpace(msh, "CG", deg)

u = TrialFunction(V)
v = TestFunction(V)

nu = Constant(0.225)

a = (inner(u.dx(0).dx(0), v.dx(0).dx(0)) +
     inner(u.dx(1).dx(1), v.dx(1).dx(1)) +
     nu*(inner(u.dx(1).dx(1), v.dx(0).dx(0)) +
         inner(u.dx(0).dx(0), v.dx(1).dx(1))) +
     2*(1 - nu)*inner(u.dx(0).dx(1), v.dx(0).dx(1)))*dx


def delta(u):
    return div(grad(u))

def dds(u, *n):
    for s in n:
        u = dot(grad(u), s)
    return u

n = FacetNormal(msh)
t = dot(as_tensor([[0, -1], [1, 0]]), n)
h = CellSize(msh)
beta1 = Constant(40)
beta2 = Constant(40)
# Clamped external boundary
a += (inner(dds(delta(u), n) - 2*(1-nu) * dds(u, t, t, n), v) +
      inner(u, dds(delta(v), n) - 2*(1-nu) * dds(v, t, t, n)) +
      beta1/h**2 * inner(u, v) -
      inner(delta(u) - 2*(1-nu)*dds(u, t, t), dds(v, n)) -
      inner(dds(u, n), delta(v) - 2*(1-nu)*dds(v, t, t)) +
      beta2/h * inner(dds(u, n), dds(v, n)))*ds(1)

mss = inner(u, v)*dx

# This prefix allows us to control factorisation package from options.
A = assemble(a, options_prefix="st_").M.handle
M = assemble(mss, options_prefix="st_").M.handle

E = SLEPc.EPS().create(comm=msh.comm)
E.setOperators(A, M)
E.setProblemType(SLEPc.EPS.ProblemType.GHEP)
# Defaults  -eps_nev 5 -eps_tol 1e-11 -eps_target 10 -st_type sinvert
E.setTolerances(tol=1e-11)
E.setTarget(10)
E.setDimensions(nev=10)
E.st.setType(SLEPc.ST.Type.SINVERT)
E.setFromOptions()
E.solve()

Print = PETSc.Sys.Print

Print()
Print("******************************")
Print("*** SLEPc Solution Results ***")
Print("******************************")
Print()

its = E.getIterationNumber()
Print("Number of iterations of the method: %d" % its)

eps_type = E.getType()
Print("Solution method: %s" % eps_type)

nev, ncv, mpd = E.getDimensions()
Print("Number of requested eigenvalues: %d" % nev)

tol, maxit = E.getTolerances()
Print("Stopping condition: tol=%.4g, maxit=%d" % (tol, maxit))

nconv = E.getConverged()
Print("Number of converged eigenpairs %d" % nconv)

ewr = Function(V)
ewi = Function(V)


fl = File("pictures/guitar/guitar.pvd")
ewrPk = Function(Pk)
vizP1 = Function(FunctionSpace(vizmesh, "CG", 1))

plotstuff = True
if plotstuff:
    import matplotlib.pyplot as plt

# Plot the domain
axes = plot(msh, surface=True)
axes.axis("off")
axes.get_xaxis().set_visible(False)
axes.get_yaxis().set_visible(False)
axes.collections[-1].set_color("black")
plt.savefig("pictures/guitar/guitar_domain.pdf", format="pdf",
            bbox_inches="tight", pad_inches=0)
plt.close()
del axes

if nconv > 0:
    vr, wr = A.getVecs()
    vi, wi = A.getVecs()
    ks = []
    for i in range(nconv):
        k = E.getEigenpair(i, vr, vi)
        ks.append(k.real)

    inds = np.argsort(ks)

triangulation = None
if nconv > 0:
    # Create the results vectors
    vr, wr = A.getVecs()
    vi, wi = A.getVecs()
    #
    Print()
    Print("        k          ||Ax-kx||/||kx||")
    Print("----------------- ------------------")
    for j, i in enumerate(inds):
        k = E.getEigenpair(i, vr, vi)
        error = E.computeError(i)
        if k.real < 1e-10:
            continue
        if k.imag != 0.0:
            Print(" %9f%+9f j %12g" % (k.real, k.imag, error))
        else:
            Print(" %12f      %12g" % (k.real, error))
        ewr.vector()[:] = vr
        ewi.vector()[:] = vi
        assert sqrt(assemble(ewi**2*dx)) < 1.e-10
        ewrPk.project(ewr)
        prolong(ewrPk, vizP1)
        fl.write(vizP1)
        if plotstuff:
            from matplotlib.tri import Triangulation
            mesh = vizP1.ufl_domain()
            fig = plt.figure()
            axes = fig.add_subplot(111)
            axes.set_aspect("equal")
            axes.axis("off")
            axes.get_xaxis().set_visible(False)
            axes.get_yaxis().set_visible(False)
            if triangulation is None:
                x = mesh.coordinates.dat.data_ro[:, 0]
                y = mesh.coordinates.dat.data_ro[:, 1]
                cells = mesh.coordinates.cell_node_map().values
                triangulation = Triangulation(x, y, cells)
            cs = axes.tricontour(triangulation,
                                 vizP1.dat.data_ro,
                                 levels=sorted([0] + list(range(-17, 18, 2))),
                                 linewidths=1, colors="k")
            axes = cs.ax
            for lev, line in zip(cs.levels, cs.collections):
                if lev == 0:
                    line.set_linewidth(1.75)
                line.set_color("gray")
            axes = plot(mesh, surface=True, axes=axes)
            axes.collections[-1].set_color("black")
            axes.collections[-1].set_linewidth(2.5)
            plt.savefig('pictures/guitar/cutaway-eigfunc%d.pdf' % (j,), format='pdf',
                        bbox_inches="tight", pad_inches=0)
            plt.close()
    Print()
