from firedrake import *
import numpy
from firedrake.slate.slac.compiler import compile_expression
from firedrake.tsfc_interface import compile_form
import os

if not os.path.exists("data"):
    os.makedirs("data")
elif not os.path.isdir("data"):
    raise RuntimeError("Cannot create output directory, file of given name exists")

msh = UnitTriangleMesh()

els = ['Lagrange']*3 + ['Hermite', 'Bell', 'Argyris']
degs = [3, 4, 5, 3, 5, 5]
names = ['{$P^%d$}' % k for k in range(3, 6)] + els[3:]


flops = []

for el, deg, name in zip(els, degs, names):
    V = FunctionSpace(msh, el, deg)

    n = FacetNormal(msh)
    h = CellSize(msh)
    beta = Constant(20)

    u = TrialFunction(V)
    v = TestFunction(V)

    a = inner(grad(u), grad(v))*dx

    # Ignore surface terms enforcing boundary conditions for flop
    # counts (killed by surface-to-volume ratios in real applications)
    # a += (-inner(dot(grad(u), n), v)*ds -
    #       inner(u, dot(grad(v), n))*ds +
    #       inner(beta/h*u, v)*ds)

    kernel, = compile_form(a, 'poisson_{}_{}'.format(el, deg))
    A_flops = kernel.kinfo.kernel.num_flops

    sparams = {
        'mat_type': 'matfree',
        'ksp_type': 'preonly',
        'pc_type': 'python',
        'pc_python_type': 'scpc.SCCG',
        'static_condensation': {
            'ksp_type': 'preonly',
            'pc_type': 'lu'}
    }
    # Slate PC to setup Schur complement static condensation.
    if el == 'Lagrange':
        problem = LinearVariationalProblem(a, 0, Function(V))
        solver = LinearVariationalSolver(problem, solver_parameters=sparams)
        solver.snes.setUp()
        solver.snes.ksp.pc.setUp()
        S = solver.snes.ksp.pc.getPythonContext().S
        slate_kernels = compile_expression(S.a)
        S_flops = sum(k.kinfo.kernel.num_flops for k in slate_kernels) - A_flops
    else:
        S_flops = 0

    flops.append([name, A_flops, S_flops])

with open('data/poissonflops.csv', 'w') as f:
    flops = numpy.asarray(flops, dtype=object)
    numpy.savetxt(f, flops, fmt=['%s', '%d', '%d'], delimiter=',',
                  header='Name, A Flops, S Flops', comments='')
