from firedrake import *
import scipy
import numpy
import matplotlib.pyplot as plt
import os

if not os.path.exists("pictures/sparsity"):
    os.makedirs("pictures/sparsity")
elif not os.path.isdir("pictures/sparsity"):
    raise RuntimeError("Cannot create output directory, file of given name exists")

N = 8
msh = UnitSquareMesh(N, N)
FONTSIZE = 18


def delta(u):
    return div(grad(u))


def dds(u, n):
    if isinstance(n, list) or isinstance(n, tuple):
        expr = u
        for s in n:
            expr = dot(grad(expr), s)
        return expr
    else:
        return dot(grad(u), n)


sparsity_stats = []
for el, deg in zip(["Morley", "Bell", "Argyris"], [2, 5, 5]):
    V = FunctionSpace(msh, el, deg)

    n = FacetNormal(msh)
    h = CellSize(msh)
    beta = Constant(20)

    n = FacetNormal(msh)
    t = dot(as_tensor([[0, -1], [1, 0]]), n)
    h = CellSize(msh)
    beta1 = Constant(40)
    beta2 = Constant(40)

    nu = Constant(0.225)

    u = TrialFunction(V)
    v = TestFunction(V)

    a = (inner(delta(u), delta(v))*dx -
         (1-nu)*(2*inner(u.dx(0).dx(0), v.dx(1).dx(1)) +
                 2*inner(u.dx(1).dx(1), v.dx(0).dx(0)) -
                 4*inner(u.dx(0).dx(1), v.dx(0).dx(1)))*dx +
         inner(dds(delta(u), n) - 2*(1-nu) * dds(u, [t, t, n]), v)*ds +
         inner(u, dds(delta(v), n) - 2*(1-nu) * dds(v, [t, t, n]))*ds +
         beta1/h**2 * inner(u, v)*ds -
         inner(delta(u) - 2*(1-nu)*dds(u, [t, t]), dds(v, n))*ds -
         inner(dds(u, n), delta(v) - 2*(1-nu)*dds(v, [t, t]))*ds +
         beta2/h * inner(dds(u, n), dds(v, n))*ds)

    A = assemble(a).M.handle
    ai, aj, av = A.getValuesCSR()
    Asp = scipy.sparse.csr_matrix((av, aj, ai))
    nrow = A.getSize()[0]

    plt.spy(Asp, markersize=0.5, aspect="equal", precision="present",
            color="black")
    ax = plt.gca()
    ax.set_xticks([0, nrow])
    ax.set_yticks([0, nrow])
    for tick in ax.get_xticklabels():
        tick.set_fontsize(FONTSIZE)
    for tick in ax.get_yticklabels():
        tick.set_fontsize(FONTSIZE)
    plt.savefig("pictures/sparsity/biharmonic-{el}-{deg}.png".format(el=el, deg=deg),
                format="png", bbox_inches="tight", pad_inches=0,
                dpi=200, background="transparent", fontsize=FONTSIZE)
    nnz = Asp.nnz
    nrows = Asp.shape[0]
    kappa = numpy.linalg.cond(Asp.todense())
    sparsity_stats.append([el, nrows, nnz, nnz/float(nrows), kappa])

with open("data/biharmonic-h2-stats.csv", "w") as f:
    numpy.savetxt(f, sparsity_stats, fmt=['%s', '%s', '%s', '%s', '%s'], delimiter=",",
                  header="Name,Nrows,NNZ,NNZperRow,kappa", comments="")
