from firedrake import *
import numpy
import scipy, scipy.sparse
import matplotlib.pyplot as plt
import os

if not os.path.exists("pictures/sparsity"):
    os.makedirs("pictures/sparsity")
elif not os.path.isdir("pictures/sparsity"):
    raise RuntimeError("Cannot create output directory, file of given name exists")

N = 8
msh = UnitSquareMesh(N, N)
FONTSIZE = 18


def delta(u):
    return div(grad(u))


def dds(u, n):
    if isinstance(n, list) or isinstance(n, tuple):
        expr = u
        for s in n:
            expr = dot(grad(expr), s)
        return expr
    else:
        return dot(grad(u), n)


sparsity_stats = []
for el, deg in zip(["Lagrange"]*3, [2, 4, 5]):
    name = '{$P^%d$}' % deg
    
    V = FunctionSpace(msh, el, deg)

    n = FacetNormal(msh)
    h = CellSize(msh)
    h_avg = (h('+') + h('-'))/2.0
    u = TrialFunction(V)
    v = TestFunction(V)

    alpha = Constant(8.0)
    u = TrialFunction(V)
    v = TestFunction(V)
    beta = Constant(40)

    a = (inner(div(grad(u)), div(grad(v)))*dx -
         inner(avg(div(grad(u))), jump(grad(v), n))*dS -
         inner(jump(grad(u), n), avg(div(grad(v))))*dS +
         alpha/h_avg*inner(jump(grad(u), n), jump(grad(v), n))*dS -
         inner(delta(u), dds(v, n))*ds - inner(dds(u, n), delta(v))*ds +
         beta/h * inner(dds(u, n), dds(v, n))*ds)

    A = assemble(a).M.handle
    nrow = A.getSize()[0]
    ai, aj, av = A.getValuesCSR()
    Asp = scipy.sparse.csr_matrix((av, aj, ai))

    plt.spy(Asp, markersize=0.5, aspect="equal", precision="present",
            color="black")
    ax = plt.gca()
    ax.set_xticks([0, nrow])
    ax.set_yticks([0, nrow])
    for tick in ax.get_xticklabels():
        tick.set_fontsize(FONTSIZE)
    for tick in ax.get_yticklabels():
        tick.set_fontsize(FONTSIZE)
    plt.savefig("pictures/sparsity/biharmonicdg-{el}-{deg}.png".format(el=el, deg=deg),
                format="png", bbox_inches="tight", pad_inches=0,
                dpi=200, background="transparent", fontsize=FONTSIZE)

    # Cut out the Dirichlet rows/columns
    bc = DirichletBC(V, 0, "on_boundary")
    nnz = Asp.nnz
    nrows = Asp.shape[0]
    Adense = numpy.asarray(Asp.todense())
    mask = numpy.zeros(nrows, dtype=bool)
    mask[bc.nodes] = True
    mask = ~mask
    Adense = Adense[numpy.ix_(mask, mask)]
    kappa = numpy.linalg.cond(Adense)
    el = "{P^%d}" % deg
    sparsity_stats.append([name, nrows, nnz, nnz/float(nrows), kappa])

with open("data/biharmonic-h1-stats.csv", "w") as f:
    numpy.savetxt(f, sparsity_stats, fmt=['%s', '%s', '%s', '%s', '%s'], delimiter=",",
                  header="Name,Nrows,NNZ,NNZperRow,kappa", comments="")
